"""
Create smooth camera paths from keyframes.



"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import random
import numpy as np
from yt.visualization.volume_rendering.create_spline import create_spline

class Keyframes(object):
    def __init__(self, x, y, z=None, north_vectors=None, up_vectors=None,
                 times=None, niter=50000, init_temp=10.0, alpha=0.999,
                 fixed_start=False):
        r"""Keyframes for camera path generation.

        From a set of keyframes with position and optional up and
        north vectors, an interpolated camera path is generated.

        Parameters
        ----------
        x : array_like
            The x positions of the keyframes
        y : array_like
            The y positions of the keyframes
        z : array_like, optional
            The z positions of the keyframes. Default: 0.0
        north_vectors : array_like, optional
            The north vectors of the keyframes. Default: None
        up_vectors : array_like, optional
            The up vectors of the keyframes. Default: None
        times : array_like, optional
            The times of the keyframes. Default: arange(N)
        niter : integer, optional
            Maximum number of iterations to find solution. Default: 50000
        init_temp : float, optional
            Initial temperature for simulated annealing when finding a
            solution.  Lower initial temperatures result in an initial solution
            in first several iterations that changes more rapidly. Default: 10.0
        alpha : float, optional
            Exponent in cooling function in simulated annealing.  Must be < 1.
            In each iteration, the temperature_new = temperature_old * alpha.
            Default: 0.999
        fixed_start: boolean, optional
            If true, the first point never changes when searching for shortest
            path.  Default: False

        Examples
        --------

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from yt.visualization.volume_rendering.camera_path import *

        # Make a camera path from 10 random (x,y,z) keyframes
        >>> data = np.random.random.((10,3))
        >>> kf = Keyframes(data[:,0], data[:,1], data[:,2])
        >>> path = kf.create_path(250, shortest_path=False)

        # Plot the keyframes in the x-y plane and camera path
        plt.plot(kf.pos[:,0], kf.pos[:,1], 'ko')
        plt.plot(path['position'][:,0], path['position'][:,1])
        plt.savefig('path.png')
        """
        Nx = len(x)
        Ny = len(y)
        if z is not None:
            Nz = len(z)
            ndims = 3
        else:
            Nz = 1
            ndims = 2
        if Nx*Ny*Nz != Nx**ndims:
            print("Need Nx (%d) == Ny (%d) == Nz (%d)" % (Nx, Ny, Nz))
            raise RuntimeError
        self.nframes = Nx
        self.pos = np.zeros((Nx,3))
        self.pos[:,0] = x
        self.pos[:,1] = y
        if z is not None:
            self.pos[:,2] = z
        else:
            self.pos[:,2] = 0.0
        self.north_vectors = north_vectors
        self.up_vectors = up_vectors
        if times is None:
            self.times = np.arange(self.nframes)
        else:
            self.times = times
        self.cartesian_matrix()
        self.setup_tsp(niter, init_temp, alpha, fixed_start)

    def setup_tsp(self, niter=50000, init_temp=10.0, alpha=0.999,
                  fixed_start=False):
        r"""Setup parameters for Travelling Salesman Problem.

        Parameters
        ----------
        niter : integer, optional
            Maximum number of iterations to find solution. Default: 50000
        init_temp : float, optional
            Initial temperature for simulated annealing when finding a
            solution.  Lower initial temperatures result in an initial solution
            in first several iterations that changes more rapidly. Default: 10.0
        alpha : float, optional
            Exponent in cooling function in simulated annealing.  Must be < 1.
            In each iteration, the temperature_new = temperature_old * alpha.
            Default: 0.999
        fixed_start: boolean, optional
            If true, the first point never changes when searching for shortest
            path.  Default: False
        """
        # randomize tour
        self.tour = range(self.nframes)
        np.random.shuffle(self.tour)
        if fixed_start:
            first = self.tour.index(0)
            self.tour[0], self.tour[first] = self.tour[first], self.tour[0]
        self.max_iterations = niter
        self.initial_temp = init_temp
        self.alpha = alpha
        self.fixed_start = fixed_start
        self.best_score = None
        self.best = None

    def set_times(self, times):
        self.times = times
        
    def rand_seq(self):
        r"""
        Generates values in random order, equivalent to using shuffle
        in random without generation all values at once.
        """
        values = range(self.nframes)
        for i in range(self.nframes):
            # pick a random index into remaining values
            j = i + int(random.random() * (self.nframes-i))
            # swap the values
            values[j], values[i] = values[i], values[j]
            # return the swapped value
            yield values[i]

    def all_pairs(self):
        r"""
        Generates all (i,j) pairs for (i,j) for 0-size
        """
        for i in self.rand_seq():
            for j in self.rand_seq():
                yield (i,j)
    
    def reversed_sections(self, tour):
        r"""
        Generator to return all possible variations where a section
        between two cities are swapped.
        """
        for i,j in self.all_pairs():
            if i == j: continue
            copy = tour[:]
            if i < j:
                copy[i:j+1] = reversed(tour[i:j+1])
            else:
                copy[i+1:] = reversed(tour[:j])
                copy[:j] = reversed(tour[i+1:])
            if self.fixed_start:
                ind = copy.index(0)
                copy[0], copy[ind] = copy[ind], copy[0]
            if copy != tour: # no point return the same tour
                yield copy

    def cartesian_matrix(self):
        r"""
        Create a distance matrix for the city coords that uses
        straight line distance
        """
        self.dist_matrix = np.zeros((self.nframes, self.nframes))
        xmat = np.zeros((self.nframes, self.nframes))
        xmat[:,:] = self.pos[:,0]
        dx = xmat - xmat.T
        ymat = np.zeros((self.nframes, self.nframes))
        ymat[:,:] = self.pos[:,1]
        dy = ymat - ymat.T
        zmat = np.zeros((self.nframes, self.nframes))
        zmat[:,:] = self.pos[:,2]
        dz = zmat - zmat.T
        self.dist_matrix = np.sqrt(dx*dx + dy*dy + dz*dz)

    def tour_length(self, tour):
        r"""
        Calculate the total length of the tour based on the distance
        matrix
        """
        total = 0
        num_cities = len(tour)
        for i in range(num_cities):
            j = (i+1) % num_cities
            city_i = tour[i]
            city_j = tour[j]
            total += self.dist_matrix[city_i, city_j]
        return -total

    def cooling(self):
        T = self.initial_temp
        while True:
            yield T
            T = self.alpha * T

    def prob(self, prev, next, temperature):
        if next > prev:
            return 1.0
        else:
            return np.exp( -abs(next-prev) / temperature )

    def get_shortest_path(self):
        r"""Determine shortest path between all keyframes.

        Parameters
        ----------
        None.
        """
        # this obviously doesn't work. When someone fixes it, remove the NOQA
        self.setup_tsp(niter, init_temp, alpha, fixed_start)  # NOQA
        num_eval = 1
        cooling_schedule = self.cooling()
        current = self.tour
        current_score = self.tour_length(current)
        for temperature in cooling_schedule:
            done = False
            # Examine moves around the current position
            for next in self.reversed_sections(current):
                if num_eval >= self.max_iterations:
                    done = True
                    break
                next_score = self.tour_length(next)
                num_eval += 1

                # Anneal.  Accept new solution if a random number is
                # greater than our "probability".
                p = self.prob(current_score, next_score, temperature)
                if random.random() < p:
                    current = next
                    self.current_score = next_score
                    if self.current_score > self.best_score:
                        #print num_eval, self.current_score, self.best_score, current
                        self.best_score = self.current_score
                        self.best = current
                    break

            if done:
                break
        self.pos = self.pos[self.tour,:]
        if self.north_vectors is not None:
            self.north_vectors = self.north_vectors[self.tour]
        if self.up_vectors is not None:
            self.up_vectors = self.up_vectors[self.tour]

    def create_path(self, npoints, path_time=None, tension=0.5, shortest_path=False):
        r"""Create a interpolated camera path from keyframes.

        Parameters
        ----------
        npoints : integer
            Number of points to interpolate from keyframes
        path_time : array_like, optional
            Times of interpolated points.  Default: Linearly spaced
        tension : float, optional
            Controls how sharp of a curve the spline takes.  A higher tension
            allows for more sharp turns.  Default: 0.5
        shortest_path : boolean, optional
            If true, estimate the shortest path between the keyframes.
            Default: False

        Returns
        -------
        path : dict
            Dictionary (time, position, north_vectors, up_vectors) of camera
            path.  Also saved to self.path.
        """
        self.npoints = npoints
        self.path = {"time": np.zeros(npoints),
                     "position": np.zeros((npoints, 3)),
                     "north_vectors": np.zeros((npoints,3)),
                     "up_vectors": np.zeros((npoints,3))}
        if shortest_path:
            self.get_shortest_path()
        if path_time is None:
            path_time = np.linspace(0, self.nframes, npoints)
        self.path["time"] = path_time
        for dim in range(3):
            self.path["position"][:,dim] = create_spline(self.times, self.pos[:,dim],
                                                         path_time, tension=tension)
            if self.north_vectors is not None:
                self.path["north_vectors"][:,dim] = \
                    create_spline(self.times, self.north_vectors[:,dim],
                                  path_time, tension=tension)
            if self.up_vectors is not None:
                self.path["up_vectors"][:,dim] = \
                    create_spline(self.times, self.up_vectors[:,dim],
                                  path_time, tension=tension)
        return self.path

    def write_path(self, filename="path.dat"):
        r"""Writes camera path to ASCII file

        Parameters
        ----------
        filename : string, optional
            Filename containing the camera path.  Default: path.dat
        """
        fp = open(filename, "w")
        fp.write("#%11s %12s %12s %12s %12s %12s %12s %12s %12s\n" %
                 ("x", "y", "z", "north_x", "north_y", "north_z",
                  "up_x", "up_y", "up_z"))
        for i in range(self.npoints):
            fp.write("%.12f %.12f %.12f %.12f %.12f %.12f %.12f %.12f %.12f\n" %
                     (self.path["position"][i,0], self.path["position"][i,1],
                      self.path["position"][i,2], 
                      self.path["north_vectors"][i,0], self.path["north_vectors"][i,1],
                      self.path["north_vectors"][i,2], 
                      self.path["up_vectors"][i,0], self.path["up_vectors"][i,1],
                      self.path["up_vectors"][i,2]))
        fp.close()




from yt.testing import \
    assert_array_equal, \
    fake_random_ds
from yt.units import cm

def test_box_creation():

    # test that creating a region with left and right edge
    # with units works
    ds = fake_random_ds(32, length_unit=2)
    reg = ds.box([0, 0, 0]*cm, [2, 2, 2]*cm)
    dens_units = reg['density']

    reg = ds.box([0, 0, 0], [1, 1, 1])
    dens_no_units = reg['density']

    assert_array_equal(dens_units, dens_no_units)



import os
"""
Compatibility module



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .grid_traversal import *
from .particle_mesh_operations import *
from .depth_first_octree import *
from .fortran_reader import *
from .interpolators import *
from .misc_utilities import *
from .basic_octree import *
from .image_utilities import *
from .points_in_volume import *
from .quad_tree import *
from .marching_cubes import *
from .write_array import *
from .mesh_utilities import *
from .contour_finding import *
from .line_integral_convolution import *



"""
This is a container for storing local fields defined on each load of yt.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
import warnings

from yt.utilities.logger import \
    ytLogger as mylog

from .field_plugin_registry import \
    register_field_plugin

from .field_info_container import \
    FieldInfoContainer

class LocalFieldInfoContainer(FieldInfoContainer):
    def add_field(self, name, function=None, sampling_type=None, **kwargs):
        if not isinstance(name, tuple):
            if kwargs.setdefault('particle_type', False):
                name = ('all', name)
            else:
                name = ('gas', name)
        override = kwargs.get("force_override", False)
        # Handle the case where the field has already been added.
        if not override and name in self:
            mylog.warning("Field %s already exists. To override use " +
                          "force_override=True.", name)
        if kwargs.setdefault('particle_type', False):
            if sampling_type is not None and sampling_type != "particle":
                raise RuntimeError("Clashing definition of 'sampling_type' and "
                               "'particle_type'. Note that 'particle_type' is "
                               "deprecated. Please just use 'sampling_type'.")
            else:
                sampling_type = "particle"
        if sampling_type is None:
            warnings.warn("Because 'sampling_type' not specified, yt will "
                          "assume a cell 'sampling_type'")
            sampling_type = "cell"
        return super(LocalFieldInfoContainer,
                     self).add_field(name, sampling_type, function, **kwargs)

# Empty FieldInfoContainer
local_fields = LocalFieldInfoContainer(None, [], None)

add_field = derived_field = local_fields.add_field

@register_field_plugin
def setup_local_fields(registry, ftype = "gas", slice_info = None):
    # This is easy.  We just update with the contents of the local_fields field
    # info container, and since they are not mutable in any real way, we are
    # fine.
    # Note that we actually don't care about the ftype here.
    for f in local_fields:
        registry._show_field_errors.append(f)
    registry.update(local_fields)



"""
Minimalist performance counting for yt



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import atexit
import time

from bisect import insort
from collections import defaultdict
from datetime import datetime as dt
from functools import wraps

from yt.config import ytcfg
from yt.funcs import mylog

class PerformanceCounters(object):
    _shared_state = {}
    def __new__(cls, *args, **kwargs):
        self = object.__new__(cls, *args, **kwargs)
        self.__dict__ = cls._shared_state
        return self

    def __init__(self):
        self.counters = defaultdict(lambda: 0.0)
        self.counting = defaultdict(lambda: False)
        self.starttime = defaultdict(lambda: 0)
        self.endtime = defaultdict(lambda: 0)
        self._on = ytcfg.getboolean("yt", "timefunctions")
        self.exit()

    def __call__(self, name):
        if not self._on: return
        if self.counting[name]:
            self.counters[name] = time.time() - self.counters[name] 
            self.counting[name] = False
            self.endtime[name] = dt.now()
        else:
            self.counters[name] = time.time()
            self.counting[name] = True
            self.starttime[name] = dt.now()

    def call_func(self, func):
        if not self._on: return func
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            self(func.__name__)
            func(*args, **kwargs)
            self(func.__name__)
        return func_wrapper

    def print_stats(self):
        mylog.info("Current counter status:\n")
        times = []
        for i in self.counters:
            insort(times, [self.starttime[i], i, 1]) # 1 for 'on'
            if not self.counting[i]:
                insort(times, [self.endtime[i], i, 0]) # 0 for 'off'
        shifts = {}
        order = []
        endtimes = {}
        shift = 0
        multi = 5
        for i in times:
            # a starting entry
            if i[2] == 1:
                shifts[i[1]] = shift
                order.append(i[1])
                shift += 1
            if i[2] == 0:
                shift -= 1
                endtimes[i[1]] = self.counters[i[1]]
        line = ''
        for i in order:
            if self.counting[i]:
                line = "%s%s%i : %s : still running\n" % (line, " "*shifts[i]*multi, shifts[i], i)
            else:
                line = "%s%s%i : %s : %0.3e\n" % (line, " "*shifts[i]*multi, shifts[i], i, self.counters[i])
        mylog.info("\n" + line)

    def exit(self):
        if self._on:
            atexit.register(self.print_stats)

yt_counters = PerformanceCounters()
time_function = yt_counters.call_func


class ProfilingController(object):
    def __init__(self):
        self.profilers = {}

    def profile_function(self, function_name):
        def wrapper(func):
            try:
                import cProfile
            except ImportError:
                return func
            my_prof = cProfile.Profile()
            self.profilers[function_name] = my_prof
            @wraps(func)
            def run_in_profiler(*args, **kwargs):
                my_prof.enable()
                func(*args, **kwargs)
                my_prof.disable()
            return run_in_profiler
        return wrapper

    def write_out(self, filename_prefix):
        if ytcfg.getboolean("yt","__parallel"):
            pfn = "%s_%03i_%03i" % (filename_prefix,
                     ytcfg.getint("yt", "__global_parallel_rank"),
                    ytcfg.getint("yt", "__global_parallel_size"))
        else:
            pfn = "%s" % (filename_prefix)
        for n, p in sorted(self.profilers.items()):
            fn = "%s_%s.cprof" % (pfn, n)
            mylog.info("Dumping %s into %s", n, fn)
            p.dump_stats(fn)



"""
API for yt.frontends.open_pmd



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
# Copyright (c) 2015, Daniel Grassinger (HZDR)
# Copyright (c) 2016, Fabian Koller (HZDR)
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from .data_structures import \
    OpenPMDDataset, \
    OpenPMDGrid, \
    OpenPMDHierarchy

from .fields import \
    OpenPMDFieldInfo

from .io import \
    IOHandlerOpenPMDHDF5

from . import \
    tests



"""
Definitions for polar coordinate systems




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .cylindrical_coordinates import CylindricalCoordinateHandler


class PolarCoordinateHandler(CylindricalCoordinateHandler):
    name = "polar"

    def __init__(self, ds, ordering = ('r', 'theta', 'z')):
        super(PolarCoordinateHandler, self).__init__(ds, ordering)
        # No need to set labels here



"""
Oct container tuned for Particles




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from oct_container cimport OctreeContainer, Oct, OctInfo, ORDER_MAX
from oct_visitors cimport cind
from libc.stdlib cimport malloc, free, qsort
from libc.math cimport floor
from yt.utilities.lib.fp_utils cimport *
cimport numpy as np
import numpy as np
from selection_routines cimport SelectorObject
cimport cython
from cython cimport floating

cdef class ParticleOctreeContainer(OctreeContainer):
    cdef Oct** oct_list
    #The starting oct index of each domain
    cdef np.int64_t *dom_offsets
    cdef public int max_level
    #How many particles do we keep befor refining
    cdef public int n_ref

    def allocate_root(self):
        cdef int i, j, k
        cdef Oct *cur
        for i in range(self.nn[0]):
            for j in range(self.nn[1]):
                for k in range(self.nn[2]):
                    cur = self.allocate_oct()
                    self.root_mesh[i][j][k] = cur

    def __dealloc__(self):
        #Call the freemem ops on every ocy
        #of the root mesh recursively
        cdef i, j, k
        if self.root_mesh == NULL: return
        for i in range(self.nn[0]):
            if self.root_mesh[i] == NULL: continue
            for j in range(self.nn[1]):
                if self.root_mesh[i][j] == NULL: continue
                for k in range(self.nn[2]):
                    if self.root_mesh[i][j][k] == NULL: continue
                    self.visit_free(self.root_mesh[i][j][k])
        free(self.oct_list)
        free(self.dom_offsets)

    cdef void visit_free(self, Oct *o):
        #Free the memory for this oct recursively
        cdef int i, j, k
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    if o.children != NULL \
                       and o.children[cind(i,j,k)] != NULL:
                        self.visit_free(o.children[cind(i,j,k)])
        free(o.children)
        free(o)

    def clear_fileind(self):
        cdef i, j, k
        for i in range(self.nn[0]):
            for j in range(self.nn[1]):
                for k in range(self.nn[2]):
                    self.visit_clear(self.root_mesh[i][j][k])

    cdef void visit_clear(self, Oct *o):
        #Free the memory for this oct recursively
        cdef int i, j, k
        o.file_ind = 0
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    if o.children != NULL \
                       and o.children[cind(i,j,k)] != NULL:
                        self.visit_clear(o.children[cind(i,j,k)])

    def __iter__(self):
        #Get the next oct, will traverse domains
        #Note that oct containers can be sorted
        #so that consecutive octs are on the same domain
        cdef int oi
        cdef Oct *o
        for oi in range(self.nocts):
            o = self.oct_list[oi]
            yield (o.file_ind, o.domain_ind, o.domain)

    def allocate_domains(self, domain_counts):
        pass

    def finalize(self):
        #This will sort the octs in the oct list
        #so that domains appear consecutively
        #And then find the oct index/offset for
        #every domain
        cdef int max_level = 0
        self.oct_list = <Oct**> malloc(sizeof(Oct*)*self.nocts)
        cdef np.int64_t i = 0, lpos = 0
        cdef int cur_dom = -1
        # Note that we now assign them in the same order they will be visited
        # by recursive visitors.
        for i in range(self.nn[0]):
            for j in range(self.nn[1]):
                for k in range(self.nn[2]):
                    self.visit_assign(self.root_mesh[i][j][k], &lpos,
                                      0, &max_level)
        assert(lpos == self.nocts)
        for i in range(self.nocts):
            self.oct_list[i].domain_ind = i
            self.oct_list[i].domain = 0
            self.oct_list[i].file_ind = -1
        self.max_level = max_level

    cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level):
        cdef int i, j, k
        self.oct_list[lpos[0]] = o
        lpos[0] += 1
        max_level[0] = imax(max_level[0], level)
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    if o.children != NULL \
                       and o.children[cind(i,j,k)] != NULL:
                        self.visit_assign(o.children[cind(i,j,k)], lpos,
                                level + 1, max_level)
        return

    cdef np.int64_t get_domain_offset(self, int domain_id):
        return 0

    cdef Oct* allocate_oct(self):
        #Allocate the memory, set to NULL or -1
        #We reserve space for n_ref particles, but keep
        #track of how many are used with np initially 0
        self.nocts += 1
        cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
        cdef int i, j, k
        my_oct.domain = -1
        my_oct.file_ind = 0
        my_oct.domain_ind = self.nocts - 1
        my_oct.children = NULL
        return my_oct

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def add(self, np.ndarray[np.uint64_t, ndim=1] indices):
        #Add this particle to the root oct
        #Then if that oct has children, add it to them recursively
        #If the child needs to be refined because of max particles, do so
        cdef np.int64_t no = indices.shape[0], p, index
        cdef int i, level
        cdef int ind[3]
        if self.root_mesh[0][0][0] == NULL: self.allocate_root()
        cdef np.uint64_t *data = <np.uint64_t *> indices.data
        cdef np.uint64_t FLAG = ~(<np.uint64_t>0)
        for p in range(no):
            # We have morton indices, which means we choose left and right by
            # looking at (MAX_ORDER - level) & with the values 1, 2, 4.
            level = 0
            index = indices[p]
            if index == FLAG:
                # This is a marker for the index not being inside the domain
                # we're interested in.
                continue
            for i in range(3):
                ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
            cur = self.root_mesh[ind[0]][ind[1]][ind[2]]
            if cur == NULL:
                raise RuntimeError
            while (cur.file_ind + 1) > self.n_ref:
                if level >= ORDER_MAX: break # Just dump it here.
                level += 1
                for i in range(3):
                    ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
                if cur.children == NULL or \
                   cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
                    cur = self.refine_oct(cur, index, level)
                    self.filter_particles(cur, data, p, level)
                else:
                    cur = cur.children[cind(ind[0],ind[1],ind[2])]
            cur.file_ind += 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef Oct *refine_oct(self, Oct *o, np.uint64_t index, int level):
        #Allocate and initialize child octs
        #Attach particles to child octs
        #Remove particles from this oct entirely
        cdef int i, j, k, m, n
        cdef int ind[3]
        cdef Oct *noct
        cdef np.uint64_t prefix1, prefix2
        # TODO: This does not need to be changed.
        o.children = <Oct **> malloc(sizeof(Oct *)*8)
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    noct = self.allocate_oct()
                    noct.domain = o.domain
                    noct.file_ind = 0
                    o.children[cind(i,j,k)] = noct
        o.file_ind = self.n_ref + 1
        for i in range(3):
            ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
        noct = o.children[cind(ind[0],ind[1],ind[2])]
        return noct

    cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,
                               int level):
        # Now we look at the last nref particles to decide where they go.
        cdef int n = imin(p, self.n_ref)
        cdef np.uint64_t *arr = data + imax(p - self.n_ref, 0)
        cdef np.uint64_t prefix1, prefix2
        # Now we figure out our prefix, which is the oct address at this level.
        # As long as we're actually in Morton order, we do not need to worry
        # about *any* of the other children of the oct.
        prefix1 = data[p] >> (ORDER_MAX - level)*3
        for i in range(n):
            prefix2 = arr[i] >> (ORDER_MAX - level)*3
            if (prefix1 == prefix2):
                o.file_ind += 1
        #print ind[0], ind[1], ind[2], o.file_ind, level

    def recursively_count(self):
        #Visit every cell, accumulate the # of cells per level
        cdef int i, j, k
        cdef np.int64_t counts[128]
        for i in range(128): counts[i] = 0
        for i in range(self.nn[0]):
            for j in range(self.nn[1]):
                for k in range(self.nn[2]):
                    if self.root_mesh[i][j][k] != NULL:
                        self.visit(self.root_mesh[i][j][k], counts)
        level_counts = {}
        for i in range(128):
            if counts[i] == 0: break
            level_counts[i] = counts[i]
        return level_counts

    cdef visit(self, Oct *o, np.int64_t *counts, level = 0):
        cdef int i, j, k
        counts[level] += 1
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    if o.children != NULL \
                       and o.children[cind(i,j,k)] != NULL:
                        self.visit(o.children[cind(i,j,k)], counts, level + 1)
        return

cdef np.uint64_t ONEBIT=1

cdef class ParticleRegions:
    cdef np.float64_t left_edge[3]
    cdef np.float64_t right_edge[3]
    cdef np.float64_t dds[3]
    cdef np.float64_t idds[3]
    cdef np.int32_t dims[3]
    cdef public np.uint64_t nfiles
    cdef public object masks

    def __init__(self, left_edge, right_edge, dims, nfiles):
        cdef int i
        self.nfiles = nfiles
        for i in range(3):
            self.left_edge[i] = left_edge[i]
            self.right_edge[i] = right_edge[i]
            self.dims[i] = dims[i]
            self.dds[i] = (right_edge[i] - left_edge[i])/dims[i]
            self.idds[i] = 1.0/self.dds[i]
        # We use 64-bit masks
        self.masks = []
        for i in range(nfiles/64 + 1):
            self.masks.append(np.zeros(dims, dtype="uint64"))

    def add_data_file(self, np.ndarray pos, int file_id, int filter = 0):
        if pos.dtype == np.float32:
            self._mask_positions[np.float32_t](pos, file_id, filter)
        elif pos.dtype == np.float64:
            self._mask_positions[np.float64_t](pos, file_id, filter)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _mask_positions(self, np.ndarray[floating, ndim=2] pos,
                              np.uint64_t file_id, int filter):
        # TODO: Replace with the bitarray
        cdef np.int64_t no = pos.shape[0]
        cdef np.int64_t p
        cdef int ind[3]
        cdef int i, use
        cdef np.ndarray[np.uint64_t, ndim=3] mask
        mask = self.masks[file_id/64]
        cdef np.uint64_t val = ONEBIT << (file_id - (file_id/64)*64)
        for p in range(no):
            # Now we locate the particle
            use = 1
            for i in range(3):
                if filter and (pos[p,i] < self.left_edge[i]
                            or pos[p,i] > self.right_edge[i]):
                    use = 0
                    break
                ind[i] = <int> ((pos[p, i] - self.left_edge[i])*self.idds[i])
                ind[i] = iclip(ind[i], 0, self.dims[i])
            if use == 1:
                mask[ind[0],ind[1],ind[2]] |= val
        return

    def identify_data_files(self, SelectorObject selector):
        # This is relatively cheap to iterate over.
        cdef int i, j, k, n
        cdef np.uint64_t fmask, offset, fcheck
        cdef np.float64_t LE[3]
        cdef np.float64_t RE[3]
        cdef np.ndarray[np.uint64_t, ndim=3] mask
        files = []
        for n in range(len(self.masks)):
            fmask = 0
            mask = self.masks[n]
            LE[0] = self.left_edge[0]
            RE[0] = LE[0] + self.dds[0]
            for i in range(self.dims[0]):
                LE[1] = self.left_edge[1]
                RE[1] = LE[1] + self.dds[1]
                for j in range(self.dims[1]):
                    LE[2] = self.left_edge[2]
                    RE[2] = LE[2] + self.dds[2]
                    for k in range(self.dims[2]):
                        if selector.select_grid(LE, RE, 0) == 1:
                            fmask |= mask[i,j,k]
                        LE[2] += self.dds[2]
                        RE[2] += self.dds[2]
                    LE[1] += self.dds[1]
                    RE[1] += self.dds[1]
                LE[0] += self.dds[0]
                RE[0] += self.dds[0]
            # Now we iterate through...
            for fcheck in range(64):
                if ((fmask >> fcheck) & ONEBIT) == ONEBIT:
                    files.append(fcheck + n * 64)
        return files




"""
Clump finder tests




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import os
import shutil
import tempfile


from yt.analysis_modules.level_sets.api import \
    Clump, \
    find_clumps, \
    get_lowest_clumps
from yt.convenience import \
    load
from yt.frontends.stream.api import \
    load_uniform_grid
from yt.testing import \
    assert_array_equal, \
    assert_equal, \
    requires_file
from yt.utilities.answer_testing.framework import \
    data_dir_load

def test_clump_finding():
    n_c = 8
    n_p = 1
    dims = (n_c, n_c, n_c)

    density = np.ones(dims)
    high_rho = 10.
    # add a couple disconnected density enhancements
    density[2, 2, 2] = high_rho
    density[6, 6, 6] = high_rho

    # put a particle at the center of one of them
    dx = 1. / n_c
    px = 2.5 * dx * np.ones(n_p)
    
    data = {"density": density,
            "particle_mass": np.ones(n_p),
            "particle_position_x": px,
            "particle_position_y": px,
            "particle_position_z": px,
            "number_of_particles": n_p}

    ds = load_uniform_grid(data, dims)

    ad = ds.all_data()
    master_clump = Clump(ad, ("gas", "density"))
    master_clump.add_validator("min_cells", 1)

    find_clumps(master_clump, 0.5, 2. * high_rho, 10.)

    # there should be two children
    assert_equal(len(master_clump.children), 2)

    leaf_clumps = get_lowest_clumps(master_clump)
    # two leaf clumps
    assert_equal(len(leaf_clumps), 2)

    # check some clump fields
    assert_equal(master_clump.children[0]["density"][0].size, 1)
    assert_equal(master_clump.children[0]["density"][0], ad["density"].max())
    assert_equal(master_clump.children[0]["particle_mass"].size, 1)
    assert_array_equal(master_clump.children[0]["particle_mass"], ad["particle_mass"])
    assert_equal(master_clump.children[1]["density"][0].size, 1)
    assert_equal(master_clump.children[1]["density"][0], ad["density"].max())
    assert_equal(master_clump.children[1]["particle_mass"].size, 0)

i30 = "IsolatedGalaxy/galaxy0030/galaxy0030"
@requires_file(i30)
def test_clump_tree_save():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    ds = data_dir_load(i30)
    data_source = ds.disk([0.5, 0.5, 0.5], [0., 0., 1.],
                          (8, 'kpc'), (1, 'kpc'))

    field = ("gas", "density")
    step = 2.0
    c_min = 10**np.floor(np.log10(data_source[field]).min()  )
    c_max = 10**np.floor(np.log10(data_source[field]).max()+1)

    master_clump = Clump(data_source, field)
    master_clump.add_info_item("center_of_mass")
    master_clump.add_validator("min_cells", 20)

    find_clumps(master_clump, c_min, c_max, step)
    leaf_clumps = get_lowest_clumps(master_clump)

    fn = master_clump.save_as_dataset(fields=["density", "x", "y", "z",
                                              "particle_mass"])
    ds2 = load(fn)

    # compare clumps in the tree
    t1 = [c for c in master_clump]
    t2 = [c for c in ds2.tree]
    mt1 = ds.arr([c.info["cell_mass"][1] for c in t1])
    mt2 = ds2.arr([c["clump", "cell_mass"] for c in t2])
    it1 = np.argsort(mt1).d.astype(int)
    it2 = np.argsort(mt2).d.astype(int)
    assert_array_equal(mt1[it1], mt2[it2])

    for i1, i2 in zip(it1, it2):
        ct1 = t1[i1]
        ct2 = t2[i2]
        assert_array_equal(ct1["gas", "density"],
                           ct2["grid", "density"])
        assert_array_equal(ct1["all", "particle_mass"],
                           ct2["all", "particle_mass"])

    # compare leaf clumps
    c1 = [c for c in leaf_clumps]
    c2 = [c for c in ds2.leaves]
    mc1 = ds.arr([c.info["cell_mass"][1] for c in c1])
    mc2 = ds2.arr([c["clump", "cell_mass"] for c in c2])
    ic1 = np.argsort(mc1).d.astype(int)
    ic2 = np.argsort(mc2).d.astype(int)
    assert_array_equal(mc1[ic1], mc2[ic2])

    os.chdir(curdir)
    shutil.rmtree(tmpdir)



from yt.funcs import issue_deprecation_warning

issue_deprecation_warning("The spectral_integrator module is deprecated. "
                          "'add_xray_emissivity_field' can now be imported "
                          "from the yt module.")

from yt.fields.xray_emission_fields import \
    add_xray_emissivity_field


"""
Boxlib frontend tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.testing import \
    assert_equal, \
    requires_file, \
    units_override_check
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    small_patch_amr, \
    data_dir_load, \
    GridValuesTest
from yt.frontends.boxlib.api import \
    OrionDataset, \
    NyxDataset, \
    WarpXDataset, \
    CastroDataset, \
    MaestroDataset
import numpy as np    

# We don't do anything needing ghost zone generation right now, because these
# are non-periodic datasets.
_orion_fields = ("temperature", "density", "velocity_magnitude")
_nyx_fields = ("Ne", "Temp", "particle_mass_density")
_warpx_fields = ("Ex", "By", "jz")
_castro_fields = ("Temp", "density", "particle_count")

radadvect = "RadAdvect/plt00000"
@requires_ds(radadvect)
def test_radadvect():
    ds = data_dir_load(radadvect)
    assert_equal(str(ds), "plt00000")
    for test in small_patch_amr(ds, _orion_fields):
        test_radadvect.__name__ = test.description
        yield test

rt = "RadTube/plt00500"
@requires_ds(rt)
def test_radtube():
    ds = data_dir_load(rt)
    assert_equal(str(ds), "plt00500")
    for test in small_patch_amr(ds, _orion_fields):
        test_radtube.__name__ = test.description
        yield test

star = "StarParticles/plrd01000"
@requires_ds(star)
def test_star():
    ds = data_dir_load(star)
    assert_equal(str(ds), "plrd01000")
    for test in small_patch_amr(ds, _orion_fields):
        test_star.__name__ = test.description
        yield test

LyA = "Nyx_LyA/plt00000"
@requires_ds(LyA)
def test_LyA():
    ds = data_dir_load(LyA)
    assert_equal(str(ds), "plt00000")
    for test in small_patch_amr(ds, _nyx_fields,
                                input_center="c",
                                input_weight="Ne"):
        test_LyA.__name__ = test.description
        yield test

@requires_file(LyA)
def test_nyx_particle_io():
    ds = data_dir_load(LyA)

    grid = ds.index.grids[0]
    npart_grid_0 = 7908  # read directly from the header
    assert_equal(grid['particle_position_x'].size, npart_grid_0)
    assert_equal(grid['DM', 'particle_position_y'].size, npart_grid_0)
    assert_equal(grid['all', 'particle_position_z'].size, npart_grid_0)

    ad = ds.all_data()
    npart = 32768  # read directly from the header
    assert_equal(ad['particle_velocity_x'].size, npart)
    assert_equal(ad['DM', 'particle_velocity_y'].size, npart)
    assert_equal(ad['all', 'particle_velocity_z'].size, npart)

    assert(np.all(ad['particle_mass'] == ad['particle_mass'][0]))

    left_edge = ds.arr([0.0, 0.0, 0.0], 'code_length')
    right_edge = ds.arr([4.0, 4.0, 4.0], 'code_length')
    center = 0.5*(left_edge + right_edge)
                   
    reg = ds.region(center, left_edge, right_edge)

    assert(np.all(np.logical_and(reg['particle_position_x'] <= right_edge[0], 
                                 reg['particle_position_x'] >= left_edge[0])))

    assert(np.all(np.logical_and(reg['particle_position_y'] <= right_edge[1], 
                                 reg['particle_position_y'] >= left_edge[1])))

    assert(np.all(np.logical_and(reg['particle_position_z'] <= right_edge[2], 
                                 reg['particle_position_z'] >= left_edge[2])))

RT_particles = "RT_particles/plt00050"
@requires_ds(RT_particles)
def test_RT_particles():
    ds = data_dir_load(RT_particles)
    assert_equal(str(ds), "plt00050")
    for test in small_patch_amr(ds, _castro_fields):
        test_RT_particles.__name__ = test.description
        yield test


@requires_file(RT_particles)
def test_castro_particle_io():
    ds = data_dir_load(RT_particles)

    grid = ds.index.grids[2]
    npart_grid_2 = 49  # read directly from the header
    assert_equal(grid['particle_position_x'].size, npart_grid_2)
    assert_equal(grid['Tracer', 'particle_position_y'].size, npart_grid_2)
    assert_equal(grid['all', 'particle_position_y'].size, npart_grid_2)

    ad = ds.all_data()
    npart = 49  # read directly from the header
    assert_equal(ad['particle_velocity_x'].size, npart)
    assert_equal(ad['Tracer', 'particle_velocity_y'].size, npart)
    assert_equal(ad['all', 'particle_velocity_y'].size, npart)

    left_edge = ds.arr([0.0, 0.0, 0.0], 'code_length')
    right_edge = ds.arr([0.25, 1.0, 1.0], 'code_length')
    center = 0.5*(left_edge + right_edge)
                   
    reg = ds.region(center, left_edge, right_edge)

    assert(np.all(np.logical_and(reg['particle_position_x'] <= right_edge[0], 
                                 reg['particle_position_x'] >= left_edge[0])))

    assert(np.all(np.logical_and(reg['particle_position_y'] <= right_edge[1], 
                                 reg['particle_position_y'] >= left_edge[1])))

langmuir = "LangmuirWave/plt00020_v2"
@requires_ds(langmuir)
def test_langmuir():
    ds = data_dir_load(langmuir)
    assert_equal(str(ds), "plt00020_v2")
    for test in small_patch_amr(ds, _warpx_fields, 
                                input_center="c",
                                input_weight="Ex"):
        test_langmuir.__name__ = test.description
        yield test

plasma = "PlasmaAcceleration/plt00030_v2"
@requires_ds(plasma)
def test_plasma():
    ds = data_dir_load(plasma)
    assert_equal(str(ds), "plt00030_v2")
    for test in small_patch_amr(ds, _warpx_fields,
                                input_center="c",
                                input_weight="Ex"):
        test_plasma.__name__ = test.description
        yield test

beam = "GaussianBeam/plt03008"
@requires_ds(beam)
def test_beam():
    ds = data_dir_load(beam)
    assert_equal(str(ds), "plt03008")
    for test in small_patch_amr(ds, _warpx_fields,
                                input_center="c",
                                input_weight="Ex"):
        test_beam.__name__ = test.description
        yield test

@requires_file(plasma)
def test_warpx_particle_io():
    ds = data_dir_load(plasma)
    grid = ds.index.grids[0]

    # read directly from the header
    npart0_grid_0 = 344  
    npart1_grid_0 = 69632

    assert_equal(grid['particle0', 'particle_position_x'].size, npart0_grid_0)
    assert_equal(grid['particle1', 'particle_position_y'].size, npart1_grid_0)
    assert_equal(grid['all', 'particle_position_z'].size, npart0_grid_0 + npart1_grid_0)

    # read directly from the header
    npart0 = 1360  
    npart1 = 802816  
    ad = ds.all_data()
    assert_equal(ad['particle0', 'particle_velocity_x'].size, npart0)
    assert_equal(ad['particle1', 'particle_velocity_y'].size, npart1)
    assert_equal(ad['all', 'particle_velocity_z'].size, npart0 + npart1)

    np.all(ad['particle1', 'particle_mass'] == ad['particle1', 'particle_mass'][0])
    np.all(ad['particle0', 'particle_mass'] == ad['particle0', 'particle_mass'][0])

    left_edge = ds.arr([-7.5e-5, -7.5e-5, -7.5e-5], 'code_length')
    right_edge = ds.arr([2.5e-5, 2.5e-5, 2.5e-5], 'code_length')
    center = 0.5*(left_edge + right_edge)
                   
    reg = ds.region(center, left_edge, right_edge)

    assert(np.all(np.logical_and(reg['particle_position_x'] <= right_edge[0], 
                                 reg['particle_position_x'] >= left_edge[0])))

    assert(np.all(np.logical_and(reg['particle_position_y'] <= right_edge[1], 
                                 reg['particle_position_y'] >= left_edge[1])))

    assert(np.all(np.logical_and(reg['particle_position_z'] <= right_edge[2], 
                                 reg['particle_position_z'] >= left_edge[2])))


_raw_fields = [('raw', 'Bx'), ('raw', 'Ey'), ('raw', 'jz')]

raw_fields = "Laser/plt00015"
@requires_ds(raw_fields)
def test_raw_fields():
    ds_fn = raw_fields
    for field in _raw_fields:
        yield GridValuesTest(ds_fn, field)


@requires_file(rt)
def test_OrionDataset():
    assert isinstance(data_dir_load(rt), OrionDataset)

@requires_file(LyA)
def test_NyxDataset():
    assert isinstance(data_dir_load(LyA), NyxDataset)

@requires_file(RT_particles)
def test_CastroDataset():
    assert isinstance(data_dir_load(RT_particles), CastroDataset)

@requires_file(LyA)
def test_WarpXDataset():
    assert isinstance(data_dir_load(plasma), WarpXDataset)

@requires_file(rt)
def test_units_override():
    units_override_check(rt)

nyx_no_particles = "nyx_sedov_plt00086"
@requires_file(nyx_no_particles)
def test_nyx_no_part():
    assert isinstance(data_dir_load(nyx_no_particles), NyxDataset)

    fields = sorted(
        [('boxlib', 'H'), ('boxlib', 'He'), ('boxlib', 'MachNumber'),
         ('boxlib', 'Ne'), ('boxlib', 'Rank'), ('boxlib', 'StateErr'),
         ('boxlib', 'Temp'), ('boxlib', 'X(H)'), ('boxlib', 'X(He)'),
         ('boxlib', 'density'), ('boxlib', 'divu'), ('boxlib', 'eint_E'),
         ('boxlib', 'eint_e'), ('boxlib', 'entropy'), ('boxlib', 'forcex'),
         ('boxlib', 'forcey'), ('boxlib', 'forcez'), ('boxlib', 'kineng'),
         ('boxlib', 'logden'), ('boxlib', 'magmom'), ('boxlib', 'magvel'),
         ('boxlib', 'magvort'), ('boxlib', 'pressure'), ('boxlib', 'rho_E'),
         ('boxlib', 'rho_H'), ('boxlib', 'rho_He'), ('boxlib', 'rho_e'),
         ('boxlib', 'soundspeed'), ('boxlib', 'x_velocity'), ('boxlib', 'xmom'),
         ('boxlib', 'y_velocity'), ('boxlib', 'ymom'), ('boxlib', 'z_velocity'),
         ('boxlib', 'zmom')])

    ds = data_dir_load(nyx_no_particles)
    assert_equal(sorted(ds.field_list), fields)

msubch = 'maestro_subCh_plt00248'
@requires_file(msubch)
def test_maestro_parameters():
    assert isinstance(data_dir_load(msubch), MaestroDataset)
    ds = data_dir_load(msubch)

    # Check a string parameter
    assert(ds.parameters['plot_base_name']=="subCh_hot_baserun_plt")
    assert(type(ds.parameters['plot_base_name']) is str)

    # Check boolean parameters: T or F
    assert(ds.parameters['use_thermal_diffusion'] is False)
    assert(type(ds.parameters['use_thermal_diffusion']) is bool)

    assert(ds.parameters['do_burning'] is True)
    assert(type(ds.parameters['do_burning']) is bool)

    # Check a float parameter with a decimal point
    assert(ds.parameters['sponge_kappa']==float('10.00000000'))
    assert(type(ds.parameters['sponge_kappa']) is float)

    # Check a float parameter with E exponent notation
    assert(ds.parameters['small_dt']==float('0.1000000000E-09'))

    # Check an int parameter
    assert(ds.parameters['s0_interp_type']==3)
    assert(type(ds.parameters['s0_interp_type']) is int)



"""
An allocation container and memory pool



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

cimport numpy as np
from libc.stdlib cimport malloc, free, realloc

cdef struct AllocationContainer:
    np.uint64_t n
    np.uint64_t n_assigned
    np.uint64_t offset
    np.int64_t con_id # container id
    void *my_objs

cdef class ObjectPool:
    cdef public np.uint64_t itemsize
    cdef np.uint64_t n_con
    cdef AllocationContainer* containers
    cdef void allocate_objs(self, int n_objs, np.int64_t con_id = ?) except *
    cdef void setup_objs(self, void *obj, np.uint64_t count, 
                         np.uint64_t offset, np.int64_t con_id)
    cdef void teardown_objs(self, void *obj, np.uint64_t n, np.uint64_t offset,
                           np.int64_t con_id)




"""
Tests for setting camera and scene attributes
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import yt.units as u

from yt.testing import \
    assert_equal, \
    fake_random_ds
from yt.visualization.volume_rendering.api import \
    Scene


valid_lens_types = [
    'plane-parallel',
    'perspective',
    'stereo-perspective',
    'fisheye',
    'spherical',
    'stereo-spherical'
]


def test_scene_and_camera_attributes():
    ds = fake_random_ds(64, length_unit=2, bbox=np.array([[-1, 1], [-1, 1], [-1, 1]]))
    sc = Scene()
    cam = sc.add_camera(ds)

    # test that initial values are correct in code units
    assert_equal(cam.width, ds.arr([3, 3, 3], 'code_length'))
    assert_equal(cam.position, ds.arr([1, 1, 1], 'code_length'))
    assert_equal(cam.focus, ds.arr([0, 0, 0], 'code_length'))

    # test setting the attributes in various ways

    attribute_values = [
        (1, ds.arr([2, 2, 2], 'code_length'), ),
        ([1], ds.arr([2, 2, 2], 'code_length'), ),
        ([1, 2], RuntimeError, ),
        ([1, 1, 1], ds.arr([2, 2, 2], 'code_length'), ),
        ((1, 'code_length'), ds.arr([1, 1, 1], 'code_length'), ),
        (((1, 'code_length'), (1, 'code_length')), RuntimeError, ),
        (((1, 'cm'), (2, 'cm'), (3, 'cm')),
         ds.arr([0.5, 1, 1.5], 'code_length'), ),
        (2*u.cm, ds.arr([1, 1, 1], 'code_length'), ),
        (ds.arr(2, 'cm'), ds.arr([1, 1, 1], 'code_length'), ),
        ([2*u.cm], ds.arr([1, 1, 1], 'code_length'), ),
        ([1, 2, 3]*u.cm, ds.arr([0.5, 1, 1.5], 'code_length'), ),
        ([1, 2]*u.cm, RuntimeError, ),
        ([u.cm*w for w in [1, 2, 3]], ds.arr([0.5, 1, 1.5], 'code_length'), ),
    ]

    # define default values to avoid accidentally setting focus = position
    default_values = {
        'focus': [0, 0, 0],
        'position': [4, 4, 4],
        'width': [1, 1, 1],
    }
    attribute_list = list(default_values.keys())

    for attribute in attribute_list:
        for other_attribute in [a for a in attribute_list if a != attribute]:
            setattr(cam, other_attribute, default_values[other_attribute])
        for attribute_value, expected_result in attribute_values:
            try:
                # test properties
                setattr(cam, attribute, attribute_value)
                assert_equal(getattr(cam, attribute), expected_result)
            except RuntimeError:
                assert expected_result is RuntimeError

            try:
                # test setters/getters
                getattr(cam, 'set_%s' % attribute)(attribute_value)
                assert_equal(getattr(cam, 'get_%s' % attribute)(),
                             expected_result)
            except RuntimeError:
                assert expected_result is RuntimeError

    resolution_values = (
        (512, (512, 512), ),
        ((512, 512), (512, 512), ),
        ((256, 512), (256, 512), ),
        ((256, 256, 256), RuntimeError),
    )

    for resolution_value, expected_result in resolution_values:
        try:
            # test properties
            cam.resolution = resolution_value
            assert_equal(cam.resolution, expected_result)
        except RuntimeError:
            assert expected_result is RuntimeError

        try:
            # test setters/getters
            cam.set_resolution(resolution_value)
            assert_equal(cam.get_resolution(), expected_result)
        except RuntimeError:
            assert expected_result is RuntimeError

    for lens_type in valid_lens_types:
        cam.set_lens(lens_type)

    # See issue #1287
    cam.focus = [0, 0, 0]
    cam_pos = [1, 0, 0]
    north_vector = [0, 1, 0]
    cam.set_position(cam_pos, north_vector)
    cam_pos = [0, 1, 0]
    north_vector = [0, 0, 1]
    cam.set_position(cam_pos, north_vector)



"""
API for yt.frontends.gamer



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
      GAMERGrid, \
      GAMERHierarchy, \
      GAMERDataset

from .fields import \
      GAMERFieldInfo

from .io import \
      IOHandlerGAMER

### NOT SUPPORTED YET
#from . import tests



"""
API for yt.analysis_modules.photon_simulator.
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.funcs import issue_deprecation_warning

issue_deprecation_warning("The photon_simulator module is deprecated. Please use pyXSIM "
                          "(http://hea-www.cfa.harvard.edu/~jzuhone/pyxsim) instead.")

from .photon_models import \
     PhotonModel, \
     ThermalPhotonModel

from .photon_simulator import \
     PhotonList, \
     EventList, \
     merge_files, \
     convert_old_file

from .spectral_models import \
     SpectralModel, \
     XSpecThermalModel, \
     XSpecAbsorbModel, \
     TableApecModel, \
     TableAbsorbModel



"""
API for yt.frontends.enzo_p



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
    EnzoPGrid, \
    EnzoPHierarchy, \
    EnzoPDataset

from .fields import \
    EnzoPFieldInfo
add_enzop_field = EnzoPFieldInfo.add_field

from .io import \
     EnzoPIOHandler

from . import tests



"""
YTData-specific fields




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


from yt.fields.field_info_container import \
    FieldInfoContainer

m_units = "g"
p_units = "cm"
v_units = "cm / s"
r_units = "cm"

class YTDataContainerFieldInfo(FieldInfoContainer):
    known_other_fields = (
    )

    known_particle_fields = (
    )

    def __init__(self, ds, field_list):
        super(YTDataContainerFieldInfo, self).__init__(ds, field_list)
        self.add_fake_grid_fields()

    def add_fake_grid_fields(self):
        """
        Add cell volume and mass fields that use the dx, dy, and dz
        fields that come with the dataset instead of the index fields
        which correspond to the oct tree.  We need to do this for now
        since we're treating the grid data like particles until we
        implement exporting AMR hierarchies.
        """

        if ("grid", "cell_volume") not in self.field_list:
            def _cell_volume(field, data):
                return data["grid", "dx"] * \
                  data["grid", "dy"] * \
                  data["grid", "dz"]
            self.add_field(("grid", "cell_volume"), sampling_type="particle",  function=_cell_volume,
                           units="cm**3")

class YTGridFieldInfo(FieldInfoContainer):
    known_other_fields = (
    )

    known_particle_fields = (
    )
